////////////////////////////////////////////////////////////////////////
// OpenTibia - an opensource roleplaying game
////////////////////////////////////////////////////////////////////////
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
////////////////////////////////////////////////////////////////////////
#include "otpch.h"
#include <iostream>

#include "tile.h"
#include "housetile.h"

#include "player.h"
#include "creature.h"

#include "teleport.h"
#include "trashholder.h"
#include "mailbox.h"

#include "combat.h"
#include "movement.h"

#include "game.h"
#include "configmanager.h"

extern ConfigManager g_config;
extern Game g_game;
extern MoveEvents* g_moveEvents;

Tile Tile::nullTile(0xFFFF, 0xFFFF, 0xFF);

bool Tile::hasProperty(enum ITEMPROPERTY prop) const
{
	if(ground && ground->hasProperty(prop))
		return true;

	if(topItems)
	{
		for(ItemVector::const_iterator it = topItems->begin(); it != topItems->end(); ++it)
		{
			if((*it)->hasProperty(prop))
				return true;
		}
	}

	if(downItems)
	{
		for(ItemVector::const_iterator it = downItems->begin(); it != downItems->end(); ++it)
		{
			if((*it)->hasProperty(prop))
				return true;
		}
	}

	return false;
}

bool Tile::hasProperty(Item* exclude, enum ITEMPROPERTY prop) const
{
	assert(exclude);
	if(ground && exclude != ground && ground->hasProperty(prop))
		return true;

	if(topItems)
	{
		for(ItemVector::const_iterator it = topItems->begin(); it != topItems->end(); ++it)
		{
			Item* item = *it;
			if(item != exclude && item->hasProperty(prop))
				return true;
		}
	}

	if(downItems)
	{
		for(ItemVector::const_iterator it = downItems->begin(); it != downItems->end(); ++it)
		{
			Item* item = *it;
			if(item != exclude && item->hasProperty(prop))
				return true;
		}
	}

	return false;
}

HouseTile* Tile::getHouseTile()
{
	if(isHouseTile())
		return static_cast<HouseTile*>(this);

	return NULL;
}

const HouseTile* Tile::getHouseTile() const
{
	if(isHouseTile())
		return static_cast<const HouseTile*>(this);

	return NULL;
}

bool Tile::isHouseTile() const
{
	return hasFlag(TILESTATE_HOUSE);
}

bool Tile::hasHeight(uint32_t n) const
{
	uint32_t height = 0;
	Item* iItem = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if((iItem = __getThing(i)->getItem()) && iItem->hasProperty(HASHEIGHT))
			++height;

		if(n == height)
			return true;
	}

	return false;
}

uint32_t Tile::getHeight() const
{
	uint32_t height = 0;
	Item* iItem = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if((iItem = __getThing(i)->getItem()) && iItem->hasProperty(HASHEIGHT))
			++height;
	}

	return height;
}

std::string Tile::getDescription(int32_t lookDistance) const
{
	return "You don't know why, but you can't see anything!";
}

Teleport* Tile::getTeleportItem() const
{
	if(!topItems)
		return NULL;

	for(ItemVector::const_iterator it = topItems->begin(); it != topItems->end(); ++it)
	{
		if((*it)->getTeleport())
			return (*it)->getTeleport();
	}

	return NULL;
}

MagicField* Tile::getFieldItem() const
{
	if(!hasFlag(TILESTATE_MAGICFIELD) || !downItems)
		return NULL;

	for(ItemVector::const_iterator it = downItems->begin(); it != downItems->end(); ++it)
	{
		if((*it)->getMagicField())
			return (*it)->getMagicField();
	}

	return NULL;
}

TrashHolder* Tile::getTrashHolder() const
{
	if(ground && ground->getTrashHolder())
		return ground->getTrashHolder();

	Item* iItem = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if((iItem = __getThing(i)->getItem()) && iItem->getTrashHolder())
			return iItem->getTrashHolder();
	}

	return NULL;
}

Mailbox* Tile::getMailbox() const
{
	if(ground && ground->getMailbox())
		return ground->getMailbox();

	Item* iItem = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if((iItem = __getThing(i)->getItem()) && iItem->getMailbox())
			return iItem->getMailbox();
	}

	return NULL;
}

BedItem* Tile::getBedItem() const
{
	if(ground && ground->getBed())
		return ground->getBed();

	Item* iItem = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if((iItem = __getThing(i)->getItem()) && iItem->getBed())
			return iItem->getBed();
	}

	return NULL;
}

Creature* Tile::getTopCreature()
{
	if(creatures && !creatures->empty())
		return *(creatures->begin());

	return NULL;
}

Item* Tile::getTopDownItem()
{
	if(downItems && !downItems->empty())
		return *(downItems->begin());

	return NULL;
}

Item* Tile::getItemByTopOrder(int32_t topOrder)
{
	if(!topItems)
		return NULL;

	//topOrder:
	//1: borders
	//2: ladders, signs, splashes
	//3: doors etc
	//4: creatures
	for(ItemVector::reverse_iterator it = topItems->rbegin(); it != topItems->rend(); ++it)
	{
		if(Item::items[(*it)->getID()].alwaysOnTopOrder == topOrder)
			return (*it);
	}

	return NULL;
}

Item* Tile::getTopTopItem()
{
	if(topItems && !topItems->empty())
		return topItems->back();

	return NULL;
}

Thing* Tile::getTopThing()
{
	Thing* thing = getTopCreature();
	if(thing)
		return thing;

	thing = getTopDownItem();
	if(thing)
		return thing;

	thing = getTopTopItem();
	if(thing)
		return thing;

	if(ground)
		return ground;

	return NULL;
}

void Tile::onAddTileItem(Item* item)
{
	updateTileFlags(item, false);

	const Position& cylinderMapPos = getPosition();
	uint32_t index = __getIndexOfThing(item);

	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);
	SpectatorVec::const_iterator it;

	CreatureVector::iterator vit;
	CreatureVector v;
	if(creatures && !creatures->empty())
	{
		for(vit = creatures->begin(); vit != creatures->end(); ++vit)
		{
			if((*vit)->isInGhostMode())
				v.push_back(*vit);
		}
	}

	//send to client
	Player* tmpPlayer = NULL;
	int32_t i = 0;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
		{
			//get the correct index
			i = index;
			if(!v.empty())
			{
				for(vit = v.begin(); vit != v.end(); ++vit)
				{
					if(!tmpPlayer->canSeeGhost(*vit))
						i--;
				}
			}

			tmpPlayer->sendAddTileItem(this, cylinderMapPos, i, item);
		}
	}

	//event methods
	for(it = list.begin(); it != list.end(); ++it)
		(*it)->onAddTileItem(this, cylinderMapPos, item);
}

void Tile::onUpdateTileItem(uint32_t index, Item* oldItem, const ItemType& oldType,
	Item* newItem, const ItemType& newType)
{
	const Position& cylinderMapPos = getPosition();

	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);
	SpectatorVec::const_iterator it;

	CreatureVector::iterator vit;
	CreatureVector v;
	if(creatures && !creatures->empty())
	{
		for(vit = creatures->begin(); vit != creatures->end(); ++vit)
		{
			if((*vit)->isInGhostMode())
				v.push_back(*vit);
		}
	}

	//send to client
	Player* tmpPlayer = NULL;
	int32_t i = 0;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
		{
			//get the correct index
			i = index;
			if(!v.empty())
			{
				for(vit = v.begin(); vit != v.end(); ++vit)
				{
					if(!tmpPlayer->canSeeGhost(*vit))
						i--;
				}
			}

			tmpPlayer->sendUpdateTileItem(this, cylinderMapPos, i, oldItem, newItem);
		}
	}

	//event methods
	for(it = list.begin(); it != list.end(); ++it)
		(*it)->onUpdateTileItem(this, cylinderMapPos, index, oldItem, oldType, newItem, newType);
}

void Tile::onRemoveTileItem(uint32_t index, Item* item)
{
	updateTileFlags(item, true);

	const Position& cylinderMapPos = getPosition();
	const ItemType& iType = Item::items[item->getID()];

	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);
	SpectatorVec::const_iterator it;

	CreatureVector::iterator vit;
	CreatureVector v;
	if(creatures && !creatures->empty())
	{
		for(vit = creatures->begin(); vit != creatures->end(); ++vit)
		{
			if((*vit)->isInGhostMode())
				v.push_back(*vit);
		}
	}

	//send to client
	Player* tmpPlayer = NULL;
	int32_t i = 0;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
		{
			//get the correct index
			i = index;
			if(!v.empty())
			{
				for(vit = v.begin(); vit != v.end(); ++vit)
				{
					if(!tmpPlayer->canSeeGhost(*vit))
						i--;
				}
			}

			tmpPlayer->sendRemoveTileItem(this, cylinderMapPos, i, item);
		}
	}

	//event methods
	for(it = list.begin(); it != list.end(); ++it)
		(*it)->onRemoveTileItem(this, cylinderMapPos, index, iType, item);
}

void Tile::onUpdateTile()
{
	const Position& cylinderMapPos = getPosition();

	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);
	SpectatorVec::const_iterator it;

	//send to client
	Player* tmpPlayer = NULL;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
			tmpPlayer->sendUpdateTile(this, cylinderMapPos);
	}

	//event methods
	for(it = list.begin(); it != list.end(); ++it)
		(*it)->onUpdateTile(this, cylinderMapPos);
}

void Tile::moveCreature(Creature* creature, Cylinder* toCylinder, bool teleport/* = false*/)
{
	int32_t oldStackPos = __getIndexOfThing(creature);
	//remove the creature
	__removeThing(creature, 0);

	Tile* toTile = toCylinder->getTile();
	// Switch the node ownership
	if(qt_node != toTile->qt_node)
	{
		qt_node->removeCreature(creature);
		toTile->qt_node->addCreature(creature);
	}

	//add the creature
	toTile->__addThing(NULL, creature);
	int32_t newStackPos = toTile->__getIndexOfThing(creature);

	Position fromPos = getPosition();
	Position toPos = toTile->getPosition();
	if(!teleport)
	{
		if(fromPos.y > toPos.y)
			creature->setDirection(NORTH);
		else if(fromPos.y < toPos.y)
			creature->setDirection(SOUTH);

		if(fromPos.x < toPos.x)
			creature->setDirection(EAST);
		else if(fromPos.x > toPos.x)
			creature->setDirection(WEST);
	}

	SpectatorVec list;
	SpectatorVec::iterator it;
	g_game.getSpectators(list, fromPos, false, true);
	g_game.getSpectators(list, toPos, true, true);

	CreatureVector::iterator vit;
	CreatureVector v;
	if(creatures && !creatures->empty())
	{
		for(vit = creatures->begin(); vit != creatures->end(); ++vit)
		{
			if((*vit)->isInGhostMode())
				v.push_back(*vit);
		}
	}

	//send to client
	Player* tmpPlayer = NULL;
	int32_t i = 0, j = 0;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()) && (!creature->isInGhostMode() || tmpPlayer->canSeeGhost(creature)))
		{
			i = oldStackPos;
			j = newStackPos;
			if(!v.empty())
			{
				for(vit = v.begin(); vit != v.end(); ++vit)
				{
					if(__getIndexOfThing(*vit) < oldStackPos)
					{
						if(!tmpPlayer->canSeeGhost(*it))
						{
							i--;
							j--;
						}
					}
				}
			}

			tmpPlayer->sendCreatureMove(creature, toTile, toPos, j, this, fromPos, i, teleport);
		}
	}

	//event method
	for(it = list.begin(); it != list.end(); ++it)
		(*it)->onCreatureMove(creature, toTile, toPos, this, fromPos, oldStackPos, teleport);

	postRemoveNotification(NULL, creature, oldStackPos, true);
	toTile->postAddNotification(NULL, creature, newStackPos);
}

ReturnValue Tile::__queryAdd(int32_t index, const Thing* thing, uint32_t count, uint32_t flags) const
{
	Thing* iithing = NULL;
	if(const Creature* creature = thing->getCreature())
	{
		if(hasBitSet(FLAG_NOLIMIT, flags))
			return RET_NOERROR;

		if(hasBitSet(FLAG_PATHFINDING, flags))
		{
			if(floorChange() || positionChange())
				return RET_NOTPOSSIBLE;
		}

		if(!ground)
			return RET_NOTPOSSIBLE;

		if(const Monster* monster = creature->getMonster())
		{
			if(hasFlag(TILESTATE_PROTECTIONZONE))
				return RET_NOTPOSSIBLE;

			if(floorChange() || positionChange())
				return RET_NOTPOSSIBLE;

			if(creatures)
			{
				if(monster->canPushCreatures() && !monster->isSummon())
				{
					Creature* creature;
					for(uint32_t i = 0; i < creatures->size(); ++i)
					{
						creature = creatures->at(i);
						if(!creature || !creature->getMonster() || !creature->isPushable() || (creature->getMonster()->isSummon()
							&& creature->getMonster()->getMaster()->getPlayer()))
							return RET_NOTPOSSIBLE;
					}
				}
				else if(!creatures->empty())
				{
					uint32_t i = creatures->size();
					for(CreatureVector::const_iterator it = creatures->begin(); it != creatures->end(); ++it)
					{
						if((*it)->isInGhostMode())
							i--;
					}

					if(i > 0)
						return RET_NOTENOUGHROOM;
				}
			}

			if(hasFlag(TILESTATE_IMMOVABLEBLOCKSOLID))
				return RET_NOTPOSSIBLE;

			if(hasBitSet(FLAG_PATHFINDING, flags) && hasFlag(TILESTATE_IMMOVABLENOFIELDBLOCKPATH))
				return RET_NOTPOSSIBLE;

			if(hasFlag(TILESTATE_BLOCKSOLID) || (hasBitSet(FLAG_PATHFINDING, flags) && hasFlag(TILESTATE_NOFIELDBLOCKPATH)))
			{
				if(!(monster->canPushItems() || hasBitSet(FLAG_IGNOREBLOCKITEM, flags)))
					return RET_NOTPOSSIBLE;
			}

			if(hasFlag(TILESTATE_MAGICFIELD))
			{
				MagicField* field = getFieldItem();
				if(field && !field->isBlocking())
				{
					CombatType_t combatType = field->getCombatType();
					//There is 3 options for a monster to enter a magic field
					//1) Monster is immune
					if(!monster->isImmune(combatType))
					{
						//1) Monster is "strong" enough to handle the damage
						//2) Monster is already afflicated by this type of condition
						if(!hasBitSet(FLAG_IGNOREFIELDDAMAGE, flags))
							return RET_NOTPOSSIBLE;

						if(!monster->canPushItems() && !monster->hasCondition(Combat::DamageToConditionType(combatType), false))
							return RET_NOTPOSSIBLE;
					}
				}
			}
		}
		else if(const Player* player = creature->getPlayer())
		{
			if(creatures && !creatures->empty() && !hasBitSet(FLAG_IGNOREBLOCKCREATURE, flags))
			{
				uint32_t i = creatures->size();
				for(CreatureVector::const_iterator it = creatures->begin(); it != creatures->end(); ++it)
				{
					if((*it) && (*it)->isInGhostMode() && !player->canSeeGhost((*it)))
						i--;
				}

				if(i > 0)
					return RET_NOTENOUGHROOM;
			}

			if(!player->getParent() && hasFlag(TILESTATE_NOLOGOUT))
			{
				//player is trying to login to a "no logout" tile
				return RET_NOTPOSSIBLE;
			}

			if(player->isPzLocked() && !player->getTile()->hasFlag(TILESTATE_PVPZONE) && hasFlag(TILESTATE_PVPZONE))
			{
				//player is trying to enter a pvp zone while being pz-locked
				return RET_PLAYERISPZLOCKEDENTERPVPZONE;
			}

			if(player->isPzLocked() && player->getTile()->hasFlag(TILESTATE_PVPZONE) && !hasFlag(TILESTATE_PVPZONE))
			{
				//player is trying to leave a pvp zone while being pz-locked
				return RET_PLAYERISPZLOCKEDLEAVEPVPZONE;
			}

			if((hasFlag(TILESTATE_NOPVPZONE) || hasFlag(TILESTATE_PROTECTIONZONE)) && player->isPzLocked())
				return RET_PLAYERISPZLOCKED;
		}
		else if(creatures && !creatures->empty() && !hasBitSet(FLAG_IGNOREBLOCKCREATURE, flags))
		{
			uint32_t i = creatures->size();
			for(CreatureVector::const_iterator it = creatures->begin(); it != creatures->end(); ++it)
			{
				if((*it)->isInGhostMode())
					i--;
			}

			if(i > 0)
				return RET_NOTENOUGHROOM;
		}

		for(uint32_t i = 0; i < getThingCount(); ++i)
		{
			if(!(iithing = __getThing(i)))
				continue;

			if(const Item* iitem = iithing->getItem())
			{
				const ItemType& iiType = Item::items[iitem->getID()];
				if(iiType.blockSolid)
				{
					if(!hasBitSet(FLAG_IGNOREBLOCKITEM, flags))
						return RET_NOTENOUGHROOM;

					if(!iiType.moveable || (iitem->isLoadedFromMap() &&
						(iitem->getUniqueId() != 0 || (iitem->getActionId() != 0 && iitem->getContainer()))))
						return RET_NOTPOSSIBLE;
				}
			}
		}
	}
	else if(const Item* item = thing->getItem())
	{
#ifdef __DEBUG__
		if(!thing->getParent() && !hasBitSet(FLAG_NOLIMIT, flags))
			std::cout << "[Notice - Tile::__queryAdd] thing->getParent() == NULL" << std::endl;

#endif
		if(hasBitSet(FLAG_NOLIMIT, flags))
			return RET_NOERROR;

		bool itemIsHangable = item->isHangable();
		if(ground == NULL && !itemIsHangable)
			return RET_NOTPOSSIBLE;

		if(creatures && !creatures->empty() && item->isBlocking() && !hasBitSet(FLAG_IGNOREBLOCKCREATURE, flags))
			return RET_NOTENOUGHROOM;

		bool hasHangable = false, supportHangable = false;
		for(uint32_t i = 0; i < getThingCount(); ++i)
		{
			if(!(iithing = __getThing(i)))
				continue;

			if(const Item* iitem = iithing->getItem())
			{
				const ItemType& iiType = Item::items[iitem->getID()];
				if(iiType.isHangable)
					hasHangable = true;

				if(iiType.isHorizontal || iiType.isVertical)
					supportHangable = true;

				if(itemIsHangable && (iiType.isHorizontal || iiType.isVertical))
					{}
				else if(iiType.blockSolid)
				{
					if(!item->isPickupable())
						return RET_NOTENOUGHROOM;

					if(iiType.allowPickupable)
						continue;

					if(!iiType.hasHeight || iiType.pickupable || iiType.isBed())
						return RET_NOTENOUGHROOM;
				}
			}
		}

		if(itemIsHangable && hasHangable && supportHangable)
			return RET_NEEDEXCHANGE;
	}

	return RET_NOERROR;
}

ReturnValue Tile::__queryMaxCount(int32_t index, const Thing* thing, uint32_t count, uint32_t& maxQueryCount,
	uint32_t flags) const
{
	maxQueryCount = std::max((uint32_t)1, count);
	return RET_NOERROR;
}

ReturnValue Tile::__queryRemove(const Thing* thing, uint32_t count, uint32_t flags) const
{
	int32_t index = __getIndexOfThing(thing);
	if(index == -1)
		return RET_NOTPOSSIBLE;

	const Item* item = thing->getItem();
	if(!item)
		return RET_NOTPOSSIBLE;

	if(count == 0 || (item->isStackable() && count > item->getItemCount()))
		return RET_NOTPOSSIBLE;

	if(item->isNotMoveable() && !hasBitSet(FLAG_IGNORENOTMOVEABLE, flags))
		return RET_NOTMOVEABLE;

	return RET_NOERROR;
}

Cylinder* Tile::__queryDestination(int32_t& index, const Thing* thing, Item** destItem, uint32_t& flags)
{
	Tile* destTile = NULL;
	*destItem = NULL;
	if(floorChangeDown())
	{
		Position pos = getTilePosition();
		pos.z += 1;
		if(Tile* downTile = g_game.getTile(pos))
		{
			if(downTile->floorChange(NORTH))
				pos.y += 1;

			if(downTile->floorChange(SOUTH))
				pos.y -= 1;

			if(downTile->floorChange(EAST))
				pos.x -= 1;

			if(downTile->floorChange(WEST))
				pos.x += 1;

			destTile = g_game.getTile(pos);
		}
	}
	else if(floorChange())
	{
		Position pos = getTilePosition();
		pos.z -= 1;
		if(floorChange(NORTH))
			pos.y--;

		if(floorChange(SOUTH))
			pos.y++;

		if(floorChange(EAST))
			pos.x++;

		if(floorChange(WEST))
			pos.x--;

		destTile = g_game.getTile(pos);
	}

	if(!destTile)
		destTile = this;
	else
		flags |= FLAG_NOLIMIT; //will ignore that there is blocking items/creatures

	if(destTile)
	{
		if(Thing* destThing = destTile->getTopDownItem())
			*destItem = destThing->getItem();
	}

	return destTile;
}

void Tile::__addThing(Creature* actor, Thing* thing)
{
	__addThing(actor, 0, thing);
}

void Tile::__addThing(Creature* actor, int32_t index, Thing* thing)
{
	if(Creature* creature = thing->getCreature())
	{
		g_game.clearSpectatorCache();
		creature->setParent(this);
		if(!creatures)
			creatures = new CreatureVector;

		creatures->insert(creatures->begin(), creature);
		++thingCount;
		return;
	}

	Item* item = thing->getItem();
	if(!item)
	{
#ifdef __DEBUG__MOVESYS__
		std::cout << "Failure: [Tile::__addThing] item == NULL" << std::endl;
		DEBUG_REPORT
#endif
		return /*RET_NOTPOSSIBLE*/;
	}

	if(g_config.getBool(ConfigManager::STORE_TRASH) && !hasFlag(TILESTATE_TRASHED))
	{
		g_game.addTrash(tilePos);
		setFlag(TILESTATE_TRASHED);
	}

	item->setParent(this);
	if(item->isGroundTile())
	{
		if(!ground)
		{
			ground = item;
			++thingCount;
			onAddTileItem(item);
		}
		else
		{
			int32_t index = __getIndexOfThing(ground);
			const ItemType& oldType = Item::items[ground->getID()];
			const ItemType& newType = Item::items[item->getID()];

			Item* oldGround = ground;
			ground->setParent(NULL);
			g_game.FreeThing(ground);

			ground = item;
			updateTileFlags(oldGround, true);
			updateTileFlags(item, false);
			onUpdateTileItem(index, oldGround, oldType, item, newType);
		}
	}
	else if(item->isAlwaysOnTop())
	{
		if(item->isSplash() && topItems)
		{
			Item* oldSplash = NULL;
			for(ItemVector::iterator it = topItems->begin(); it != topItems->end(); ++it)
			{
				if((*it)->isSplash() && (oldSplash = (*it)))
				{
					__removeThing(oldSplash, 1);
					oldSplash->setParent(NULL);

					g_game.FreeThing(oldSplash);
					break;
				}
			}
		}

		bool isInserted = false;
		if(topItems)
		{
			for(ItemVector::iterator it = topItems->begin(); it != topItems->end(); ++it)
			{
				//Note: this is different from internalAddThing
				if((*it) && Item::items[item->getID()].alwaysOnTopOrder <= Item::items[(*it)->getID()].alwaysOnTopOrder)
				{
					topItems->insert(it, item);
					++thingCount;

					isInserted = true;
					break;
				}
			}
		}

		if(!isInserted)
		{
			if(!topItems)
				topItems = new ItemVector;

			topItems->push_back(item);
			++thingCount;
		}

		onAddTileItem(item);
	}
	else
	{
		if(item->isMagicField() && downItems)
		{
			//remove old field item if exists
			MagicField* oldField = NULL;
			for(ItemVector::iterator it = downItems->begin(); it != downItems->end(); ++it)
			{
				if(!(oldField = (*it)->getMagicField()))
					continue;

				if(oldField->isReplaceable())
				{
					__removeThing(oldField, 1);
					oldField->setParent(NULL);

					g_game.FreeThing(oldField);
					break;
				}
				else
				{
					//This magic field cannot be replaced.
					item->setParent(NULL);
					g_game.FreeThing(item);
					return;
				}
			}
		}

		if(!downItems)
			downItems = new ItemVector;

		downItems->insert(downItems->begin(), item);
		++thingCount;
		onAddTileItem(item);
	}
}

void Tile::__updateThing(Thing* thing, uint16_t itemId, uint32_t count)
{
	int32_t index = __getIndexOfThing(thing);
	if(index == -1)
	{
#ifdef __DEBUG__MOVESYS__
		std::cout << "Failure: [Tile::__updateThing] index == -1" << std::endl;
		DEBUG_REPORT
#endif
		return /*RET_NOTPOSSIBLE*/;
	}

	Item* item = thing->getItem();
	if(!item)
	{
#ifdef __DEBUG__MOVESYS__
		std::cout << "Failure: [Tile::__updateThing] item == NULL" << std::endl;
		DEBUG_REPORT
#endif
		return /*RET_NOTPOSSIBLE*/;
	}

	const ItemType& oldType = Item::items[item->getID()];
	const ItemType& newType = Item::items[itemId];

	updateTileFlags(item, true);
	item->setID(itemId);
	item->setSubType(count);

	updateTileFlags(item, false);
	onUpdateTileItem(index, item, oldType, item, newType);
}

void Tile::__replaceThing(uint32_t index, Thing* thing)
{
	int32_t pos = index;
	Item* item = thing->getItem();
	if(!item)
	{
#ifdef __DEBUG__MOVESYS__
		std::cout << "Failure: [Tile::__updateThing] item == NULL" << std::endl;
		DEBUG_REPORT
#endif
		return /*RET_NOTPOSSIBLE*/;
	}

	Item* oldItem = NULL;
	if(ground)
	{
		if(pos == 0)
		{
			oldItem = ground;
			ground = item;
		}

		--pos;
	}

	if(topItems)
	{
		if(!oldItem && pos < (int32_t)topItems->size())
		{
			ItemVector::iterator it = topItems->begin();
			it += pos;
			pos = 0;

			oldItem = (*it);
			it = topItems->erase(it);
			topItems->insert(it, item);
		}

		pos -= (uint32_t)topItems->size();
	}

	if(creatures)
	{
		if(!oldItem && pos < (int32_t)creatures->size())
		{
#ifdef __DEBUG__MOVESYS__
			std::cout << "Failure: [Tile::__updateThing] Update object is a creature" << std::endl;
			DEBUG_REPORT
#endif
			return /*RET_NOTPOSSIBLE*/;
		}

		pos -= (uint32_t)creatures->size();
	}

	if(downItems && !oldItem && pos < (int32_t)downItems->size())
	{
		ItemVector::iterator it = downItems->begin();
		it += pos;
		pos = 0;

		oldItem = (*it);
		it = downItems->erase(it);
		downItems->insert(it, item);
	}

	if(oldItem)
	{
		item->setParent(this);
		const ItemType& oldType = Item::items[oldItem->getID()];
		const ItemType& newType = Item::items[item->getID()];

		updateTileFlags(oldItem, true);
		updateTileFlags(item, false);
		onUpdateTileItem(index, oldItem, oldType, item, newType);

		oldItem->setParent(NULL);
		return /*RET_NOERROR*/;
	}

#ifdef __DEBUG__MOVESYS__
	std::cout << "Failure: [Tile::__updateThing] Update object not found" << std::endl;
	DEBUG_REPORT
#endif
}

void Tile::__removeThing(Thing* thing, uint32_t count)
{
	if(thing->getCreature())
	{
		if(creatures)
		{
			CreatureVector::iterator it = std::find(creatures->begin(), creatures->end(), thing);
			if(it == creatures->end())
			{
#ifdef __DEBUG__MOVESYS__
				std::cout << "Failure: [Tile::__removeThing] creature not found" << std::endl;
				DEBUG_REPORT
#endif
				return;
			}

			g_game.clearSpectatorCache();
			creatures->erase(it);
			if(creatures->empty())
			{
				delete creatures;
				creatures = NULL;
			}

			--thingCount;
			return;
		}
		else
		{
#ifdef __DEBUG__MOVESYS__
			std::cout << "Failure: [Tile::__removeThing] creature not found" << std::endl;
			DEBUG_REPORT
#endif
			return; //RET_NOTPOSSIBLE;
		}
	}
	else
	{
		Item* item = thing->getItem();
		if(!item)
		{
#ifdef __DEBUG__MOVESYS__
			std::cout << "Failure: [Tile::__removeThing] item == NULL" << std::endl;
			DEBUG_REPORT
#endif
			return;
		}

		int32_t index = __getIndexOfThing(item);
		if(index == -1)
		{
#ifdef __DEBUG__MOVESYS__
			std::cout << "Failure: [Tile::__removeThing] index == -1" << std::endl;
			DEBUG_REPORT
#endif
			return;
		}

		if(item == ground)
		{
			ground->setParent(NULL);
			ground = NULL;
			--thingCount;

			onRemoveTileItem(index, item);
			return;
		}

		if(item->isAlwaysOnTop())
		{
			if(topItems)
			{
				for(ItemVector::iterator it = topItems->begin(); it != topItems->end(); ++it)
				{
					if(*it != item)
						continue;

					(*it)->setParent(NULL);
					topItems->erase(it);
					if(topItems->empty())
					{
						delete topItems;
						topItems = NULL;
					}

					--thingCount;
					onRemoveTileItem(index, item);
					return;
				}
			}
		}
		else if(downItems)
		{
			for(ItemVector::iterator it = downItems->begin(); it != downItems->end(); ++it)
			{
				if(*it != item)
					continue;

				if(item->isStackable() && count != item->getItemCount())
				{
					updateTileFlags(item, true);
					item->setItemCount(std::max(0, (int32_t)(item->getItemCount() - count)));
						const ItemType& it = Item::items[item->getID()];

					updateTileFlags(item, false);
					onUpdateTileItem(index, item, it, item, it);
				}
				else
				{
					(*it)->setParent(NULL);
					downItems->erase(it);
					if(downItems->empty())
					{
						delete downItems;
						downItems = NULL;
					}

					--thingCount;
					onRemoveTileItem(index, item);
					return;
				}
			}
		}
	}
#ifdef __DEBUG__MOVESYS__

	std::cout << "Failure: [Tile::__removeThing] thing not found" << std::endl;
	DEBUG_REPORT
#endif
}

int32_t Tile::__getIndexOfThing(const Thing* thing) const
{
	int32_t n = -1;
	if(ground)
	{
		if(ground == thing)
			return 0;

		++n;
	}

	if(topItems)
	{
		for(ItemVector::iterator it = topItems->begin(); it != topItems->end(); ++it)
		{
			++n;
			if((*it) == thing)
				return n;
		}
	}

	if(creatures)
	{
		for(CreatureVector::const_iterator cit = creatures->begin(); cit != creatures->end(); ++cit)
		{
			++n;
			if((*cit) == thing)
				return n;
		}
	}

	if(downItems)
	{
		for(ItemVector::iterator it = downItems->begin(); it != downItems->end(); ++it)
		{
			++n;
			if((*it) == thing)
				return n;
		}
	}

	return -1;
}

int32_t Tile::__getFirstIndex() const
{
	return 0;
}

int32_t Tile::__getLastIndex() const
{
	return getThingCount();
}

uint32_t Tile::__getItemTypeCount(uint16_t itemId, int32_t subType /*= -1*/, bool itemCount /*= true*/) const
{
	uint32_t count = 0;
	Thing* thing = NULL;
	for(uint32_t i = 0; i < getThingCount(); ++i)
	{
		if(!(thing = __getThing(i)))
			continue;

		if(const Item* item = thing->getItem())
		{
			if(item->getID() == itemId && (subType == -1 || subType == item->getSubType()))
			{
				if(itemCount)
					count += item->getItemCount();
				else if(item->isRune())
					count += item->getCharges();
				else
					count += item->getItemCount();
			}
		}
	}

	return count;
}

Thing* Tile::__getThing(uint32_t index) const
{
	if(ground)
	{
		if(index == 0)
			return ground;

		--index;
	}

	if(topItems)
	{
		if((unsigned)index < topItems->size())
			return topItems->at(index);

		index -= (uint32_t)topItems->size();
	}

	if(creatures)
	{
		if((unsigned)index < creatures->size())
			return creatures->at(index);

		index -= (uint32_t)creatures->size();
	}

	if(downItems)
	{
		if((unsigned)index < downItems->size())
			return downItems->at(index);
	}

	return NULL;
}

void Tile::postAddNotification(Creature* actor, Thing* thing, int32_t index, cylinderlink_t link /*= LINK_OWNER*/)
{
	const Position& cylinderMapPos = getPosition();
	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);
	SpectatorVec::const_iterator it;

	Player* tmpPlayer = NULL;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
			tmpPlayer->postAddNotification(actor, thing, index, LINK_NEAR);
	}

	//add a reference to this item, it may be deleted after being added (mailbox for example)
	thing->useThing2();
	Item* item = thing->getItem();
	if(link == LINK_OWNER)
	{
		//calling movement scripts
		if(Creature* creature = thing->getCreature())
			g_moveEvents->onCreatureMove(creature, this, true);
		else if(item)
			g_moveEvents->onItemMove(actor, item, this, true);

		if(Teleport* teleport = getTeleportItem())
			teleport->__addThing(actor, thing);
		else if(TrashHolder* trashHolder = getTrashHolder())
			trashHolder->__addThing(actor, thing);
		else if(Mailbox* mailbox = getMailbox())
			mailbox->__addThing(actor, thing);
	}

	//release the reference to this item onces we are finished
	g_game.FreeThing(thing);
}

void Tile::postRemoveNotification(Creature* actor, Thing* thing, int32_t index, bool isCompleteRemoval, cylinderlink_t link /*= LINK_OWNER*/)
{
	const Position& cylinderMapPos = getPosition();
	const SpectatorVec& list = g_game.getSpectators(cylinderMapPos);

	SpectatorVec::const_iterator it;
	if(/*isCompleteRemoval && */getThingCount() > 8)
		onUpdateTile();

	Player* tmpPlayer = NULL;
	for(it = list.begin(); it != list.end(); ++it)
	{
		if((tmpPlayer = (*it)->getPlayer()))
			tmpPlayer->postRemoveNotification(actor, thing, index, isCompleteRemoval, LINK_NEAR);
	}

	//calling movement scripts
	Item* item = thing->getItem();
	if(Creature* creature = thing->getCreature())
		g_moveEvents->onCreatureMove(creature, this, false);
	else if(item)
		g_moveEvents->onItemMove(actor, item, this, false);
}

void Tile::__internalAddThing(Thing* thing)
{
	__internalAddThing(0, thing);
}

void Tile::__internalAddThing(uint32_t index, Thing* thing)
{
	thing->setParent(this);
	if(Creature* creature = thing->getCreature())
	{
		g_game.clearSpectatorCache();
		if(!creatures)
			creatures = new CreatureVector;

		creatures->insert(creatures->begin(), creature);
		++thingCount;
		return;
	}

	Item* item = thing->getItem();
	if(!item)
		return;

	if(item->isGroundTile())
	{
		if(ground == NULL)
		{
			ground = item;
			++thingCount;
		}
	}
	else if(item->isAlwaysOnTop())
	{
		bool isInserted = false;
		if(topItems)
		{
			for(ItemVector::iterator it = topItems->begin(); it != topItems->end(); ++it)
			{
				if(Item::items[(*it)->getID()].alwaysOnTopOrder > Item::items[item->getID()].alwaysOnTopOrder)
				{
					topItems->insert(it, item);
					++thingCount;

					isInserted = true;
					break;
				}
			}
		}

		if(!isInserted)
		{
			if(!topItems)
				topItems = new ItemVector;

			topItems->push_back(item);
			++thingCount;
		}
	}
	else
	{
		if(downItems && !downItems->empty())
			downItems->insert(downItems->begin(), item);
		else
		{
			if(!downItems)
				downItems = new ItemVector;

			downItems->push_back(item);
		}

		++thingCount;
	}

	updateTileFlags(item, false);
}

void Tile::updateTileFlags(Item* item, bool removing)
{
	if(!removing)
	{
		//!removing is adding an item to the tile
		if(!hasFlag(TILESTATE_FLOORCHANGE))
		{
			if(item->floorChangeDown())
			{
				setFlag(TILESTATE_FLOORCHANGE);
				setFlag(TILESTATE_FLOORCHANGE_DOWN);
			}

			if(item->floorChangeNorth())
			{
				setFlag(TILESTATE_FLOORCHANGE);
				setFlag(TILESTATE_FLOORCHANGE_NORTH);
			}

			if(item->floorChangeSouth())
			{
				setFlag(TILESTATE_FLOORCHANGE);
				setFlag(TILESTATE_FLOORCHANGE_SOUTH);
			}

			if(item->floorChangeEast())
			{
				setFlag(TILESTATE_FLOORCHANGE);
				setFlag(TILESTATE_FLOORCHANGE_EAST);
			}

			if(item->floorChangeWest())
			{
				setFlag(TILESTATE_FLOORCHANGE);
				setFlag(TILESTATE_FLOORCHANGE_WEST);
			}
		}

		if(item->getTeleport())
			setFlag(TILESTATE_POSITIONCHANGE);

		if(item->getMagicField())
			setFlag(TILESTATE_MAGICFIELD);

		if(item->hasProperty(BLOCKSOLID))
			setFlag(TILESTATE_BLOCKSOLID);

		if(item->hasProperty(IMMOVABLEBLOCKSOLID))
			setFlag(TILESTATE_IMMOVABLEBLOCKSOLID);

		if(item->hasProperty(BLOCKPATH))
			setFlag(TILESTATE_BLOCKPATH);

		if(item->hasProperty(NOFIELDBLOCKPATH))
			setFlag(TILESTATE_NOFIELDBLOCKPATH);

		if(item->hasProperty(IMMOVABLENOFIELDBLOCKPATH))
			setFlag(TILESTATE_IMMOVABLENOFIELDBLOCKPATH);
	}
	else
	{
		if(item->floorChangeDown())
		{
			resetFlag(TILESTATE_FLOORCHANGE);
			resetFlag(TILESTATE_FLOORCHANGE_DOWN);
		}

		if(item->floorChangeNorth())
		{
			resetFlag(TILESTATE_FLOORCHANGE);
			resetFlag(TILESTATE_FLOORCHANGE_NORTH);
		}

		if(item->floorChangeSouth())
		{
			resetFlag(TILESTATE_FLOORCHANGE);
			resetFlag(TILESTATE_FLOORCHANGE_SOUTH);
		}

		if(item->floorChangeEast())
		{
			resetFlag(TILESTATE_FLOORCHANGE);
			resetFlag(TILESTATE_FLOORCHANGE_EAST);
		}

		if(item->floorChangeWest())
		{
			resetFlag(TILESTATE_FLOORCHANGE);
			resetFlag(TILESTATE_FLOORCHANGE_WEST);
		}

		if(item->getTeleport())
			resetFlag(TILESTATE_POSITIONCHANGE);

		if(item->getMagicField())
			resetFlag(TILESTATE_MAGICFIELD);

		if(item->hasProperty(BLOCKSOLID) && !hasProperty(item, BLOCKSOLID))
			resetFlag(TILESTATE_BLOCKSOLID);

		if(item->hasProperty(IMMOVABLEBLOCKSOLID) && !hasProperty(item, IMMOVABLEBLOCKSOLID))
			resetFlag(TILESTATE_IMMOVABLEBLOCKSOLID);

		if(item->hasProperty(BLOCKPATH) && !hasProperty(item, BLOCKPATH))
			resetFlag(TILESTATE_BLOCKPATH);

		if(item->hasProperty(NOFIELDBLOCKPATH) && !hasProperty(item, NOFIELDBLOCKPATH))
			resetFlag(TILESTATE_NOFIELDBLOCKPATH);

		if(item->hasProperty(IMMOVABLEBLOCKPATH) && !hasProperty(item, IMMOVABLEBLOCKPATH))
			resetFlag(TILESTATE_IMMOVABLEBLOCKPATH);

		if(item->hasProperty(IMMOVABLENOFIELDBLOCKPATH) && !hasProperty(item, IMMOVABLENOFIELDBLOCKPATH))
			resetFlag(TILESTATE_IMMOVABLENOFIELDBLOCKPATH);
	}
}
